"""Methods for dealing with OCF lab hours.

All times are assumed to be OST (OCF Standard Time).

Usage:

    >>> from ocflib.lab.hours import Day
    >>> Day.from_date(date(2015, 10, 12))
    Day(
        date=datetime.date(2015, 10, 12),
        weekday='Monday',
        holiday=None,
        hours=[Hour(open=9, close=21)],
    )
"""
from collections import defaultdict
from collections import namedtuple
from datetime import date
from datetime import datetime
from datetime import time


class Hour(namedtuple('Hours', ['open', 'close'])):

    def __contains__(self, when):
        return self.open <= when.time() < self.close


class Day(namedtuple('Day', ['date', 'weekday', 'holiday', 'hours'])):

    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6

    @classmethod
    def from_date(cls, when=None):
        """Return whether a Day representing the given day.

        If not provided, when defaults to today.
        """
        if not when:
            when = date.today()

        if isinstance(when, datetime):
            when = when.date()

        # check if it's a holiday
        my_holiday = None
        my_hours = REGULAR_HOURS[when.weekday()]

        for start, end, name, hours in HOLIDAYS:
            if start <= when <= end:
                my_holiday = name
                my_hours = hours
                break

        return cls(
            date=when,
            weekday=when.strftime('%A'),
            holiday=my_holiday,
            hours=my_hours,
        )

    def is_open(self, when=None):
        """Return whether the lab is open at the given time.

        If not provided, when defaults to now.
        """
        if not when:
            when = datetime.now()

        if not isinstance(when, datetime):
            raise ValueError('{} must be a datetime instance'.format(when))

        return any(when in hour for hour in self.hours)

    @property
    def closed_all_day(self):
        return not self.hours


REGULAR_HOURS = defaultdict(lambda: [Hour(time(9), time(20))], {
    Day.MONDAY: [Hour(time(9), time(18))],
    Day.TUESDAY: [Hour(time(9), time(18))],
    Day.THURSDAY: [Hour(time(9), time(18))],
    Day.FRIDAY: [Hour(time(10), time(17))],
    Day.SATURDAY: [],
    Day.SUNDAY: [Hour(time(12), time(17))],
})

HOLIDAYS = [
    # start date, end date, holiday name, list of hours (date ranges are inclusive)
    (date(2016, 5, 14), date(2016, 8, 23), 'Summer Break', []),
    (date(2016, 9, 5), date(2016, 9, 5), 'Labor Day', []),
    (date(2016, 9, 7), date(2016, 9, 7), 'General Meeting', [Hour(time(9), time(18))]),
    (date(2016, 11, 11), date(2016, 11, 11), 'Veteran\'s Day', []),
    (date(2016, 11, 23), date(2016, 11, 27), 'Thanksgiving Break', []),
    (date(2016, 12, 16), date(2016, 12, 16), 'Last Day of Fall 2016', [Hour(time(9), time(12))]),
    (date(2016, 12, 17), date(2017, 1, 16), 'Winter Break', []),
]



import pytest

from ocflib.misc.validators import email_host_exists
from ocflib.misc.validators import host_exists
from ocflib.misc.validators import valid_email
from ocflib.misc.validators import valid_login_shell


REAL_HOSTS = [
    'ocf.berkeley.edu',
    'google.com',
    'dev-ocf.berkeley.edu',
    'mirrors.berkeley.edu',
    'cs.berkeley.edu',
    'cory.eecs.berkeley.edu',
    'g.berkeley.edu',
]
FAKE_HOSTS = [
    'hello.waffle',
    'i am not a real host',
    'lashdfkjashgklhsadfsad.com',
    'asdfghjkl.eecs.berkeley.edu',
    'asdfghjkl.berkeley.edu',
    'kljasdlgjlsafdfhsadf.berkeley.edu',
    'jf0194y89v(*#14o1i9XC',
    '@I$)!($U)!#Y%!)#()*(%!#',
    'vns;alf iashf poasf bawen svn;',
]


@pytest.mark.parametrize(
    'host,exists',
    [(host, True) for host in REAL_HOSTS] +
    [(host, False) for host in FAKE_HOSTS]
)
def test_host_exists(host, exists):
    assert host_exists(host) == exists


@pytest.mark.parametrize(
    'email,exists',
    [('ckuehl@' + host, True) for host in REAL_HOSTS] +
    [('ckuehl@' + host, False) for host in FAKE_HOSTS] +
    [('no host here!', False)],
)
def test_email_host_exists(email, exists):
    assert email_host_exists(email) == exists


@pytest.mark.parametrize('email,valid', [
    ('ckuehl@ocf.berkeley.edu', True),
    ('hello-world_i+am.email@google.com', True),
    ('hello world@ocf.berkeley.edu', False),
    ('hello world@ocf', False),
    ('derp@langasdgkjsadhglkjbjgsdfgsd.com', False),
    ('derp@www.ocf.berkeley.edu', False),  # no MX records
    ('@ocf', False),
    ('hello@', False),
    ('@', False),
    ('', False),
])
def test_valid_email(email, valid):
    assert valid_email(email) == valid


@pytest.mark.parametrize('shell,valid', [
    ('/bin/sh', True),
    ('/bin/bash', True),
    ('/bin/dash', True),
    ('/bin/trash', False),
    ('/bin/rbash', True),
    ('/bin/tcsh', True),
    ('/bin/zsh', True),
    ('/usr/bin/screen', True),
    ('/usr/bin/tmux', True),
    ('bash', False),
])
def test_valid_login_shell(shell, valid):
    assert valid_login_shell(shell) == valid



import crypt

import mock
import pkg_resources
import pytest

import ocflib.vhost.mail
from ocflib.vhost.mail import crypt_password
from ocflib.vhost.mail import get_mail_vhosts
from ocflib.vhost.mail import MailForwardingAddress
from ocflib.vhost.mail import MailVirtualHost
from ocflib.vhost.mail import vhosts_for_user


@pytest.fixture
def example_vhost():
    return MailVirtualHost('staff', 'dev-vhost.ocf.berkeley.edu')


@pytest.fixture
def another_example_vhost():
    return MailVirtualHost('staff', 'vhost.berkeley.edu')


@pytest.yield_fixture
def fake_vhosts():
    text = (
        '# this is a comment at the start     \n',
        '  # this is a comment with leading whitespace\n',
        '\n',
        '\n',
        '           \n',
        'staff dev-vhost.ocf.berkeley.edu  \n',
        'staff    dev-dev-vhost.ocf.berkeley.edu\n',
        'ckuehl\tckuehl.berkeley.edu\n',
        '\n',
        '# sup another comment\n',
        'daradib   \t daradib.wat \n',
        '\n',
    )
    with mock.patch.object(ocflib.vhost.mail, 'get_mail_vhost_db', return_value=text):
        yield


@pytest.mark.usefixtures('fake_vhosts')
def test_get_mail_vhosts():
    assert get_mail_vhosts() == {
        MailVirtualHost('staff', 'dev-vhost.ocf.berkeley.edu'),
        MailVirtualHost('staff', 'dev-dev-vhost.ocf.berkeley.edu'),
        MailVirtualHost('ckuehl', 'ckuehl.berkeley.edu'),
        MailVirtualHost('daradib', 'daradib.wat'),
    }


@pytest.mark.usefixtures('fake_vhosts')
def test_vhosts_for_user():
    assert vhosts_for_user('staff') == {
        MailVirtualHost('staff', 'dev-vhost.ocf.berkeley.edu'),
        MailVirtualHost('staff', 'dev-dev-vhost.ocf.berkeley.edu'),
    }
    assert vhosts_for_user('ckuehl') == {
        MailVirtualHost('ckuehl', 'ckuehl.berkeley.edu'),
    }
    assert vhosts_for_user('jvperrin') == set()


def _normalize_times(addrs):
    return {
        addr._replace(last_updated=None)
        for addr in addrs
    }


@pytest.mark.parametrize(('forward_to', 'expected'), (
    ('a@a.com,b@b.com', {'a@a.com', 'b@b.com'}),
    (' a@a.com  , b@b.com ', {'a@a.com', 'b@b.com'}),
    (', a@a.com  ,', {'a@a.com'}),
    ('', set()),
    (' ', set()),
    (' , , ', set()),
))
def test_get_forwarding_addresses_forward_to_parsing(
        forward_to,
        expected,
        example_vhost,
):
    """Make sure we're a little tolerant in forward_to parsing."""
    fake_connection = mock.Mock(
        __iter__=lambda _: iter(({
            'address': 'test@test.com',
            'password': None,
            'forward_to': forward_to,
            'last_updated': None,
        },))
    )
    result, = example_vhost.get_forwarding_addresses(fake_connection)
    assert result.forward_to == expected


def test_add_and_remove_addresses(
        mysql_connection,
        example_vhost,
        another_example_vhost,
):
    assert example_vhost.get_forwarding_addresses(mysql_connection) == set()

    addr1 = MailForwardingAddress(
        'ckuehl@dev-vhost.ocf.berkeley.edu',
        'hunter2',
        frozenset(['ckuehl@ocf.berkeley.edu']),
        None,
    )
    addr2 = MailForwardingAddress(
        '@dev-vhost.ocf.berkeley.edu',
        None,
        frozenset(['ckuehl@ocf.berkeley.edu', 'daradib@ocf.berkeley.edu']),
        None,
    )

    # add one-by-one
    example_vhost.add_forwarding_address(mysql_connection, addr1)
    assert _normalize_times(example_vhost.get_forwarding_addresses(mysql_connection)) == {addr1}

    example_vhost.add_forwarding_address(mysql_connection, addr2)
    assert _normalize_times(example_vhost.get_forwarding_addresses(mysql_connection)) == {addr1, addr2}

    # unrelated vhost should have no addresses
    assert another_example_vhost.get_forwarding_addresses(mysql_connection) == set()

    # now remove them
    example_vhost.remove_forwarding_address(
        mysql_connection,
        'ckuehl@dev-vhost.ocf.berkeley.edu',
    )
    assert _normalize_times(example_vhost.get_forwarding_addresses(mysql_connection)) == {addr2}

    example_vhost.remove_forwarding_address(
        mysql_connection,
        '@dev-vhost.ocf.berkeley.edu',
    )
    assert example_vhost.get_forwarding_addresses(mysql_connection) == set()


@pytest.mark.parametrize('password', [
    '',
    'correct horse battery staple',
    'hunter2',
    '12yuCV*()AYY!@)+R*_yf9-()@#$)_&!YU($12\'"!#$!@)[]}\\',
])
def test_crypt_password(password):
    crypted = crypt_password(password)

    # it should be a sha512 hash
    assert crypted.startswith('$6$')

    # verify password against hash succeeds
    assert crypt.crypt(password, crypted) == crypted

    # verify not-the-password against hash fails
    for not_password in ['', password + ' ', 'hunter3']:
        assert crypt.crypt(password, crypt_password(not_password)) != crypted


@pytest.yield_fixture
def mysql_connection(mysql_database):
    mysql_database.run_cli_query(
        pkg_resources.resource_string('ocflib.vhost', 'ocfmail.sql'),
    )
    with mysql_database.connection() as c:
        yield c



import pytest

from ocflib.account.search import user_attrs
from ocflib.account.search import user_attrs_ucb
from ocflib.account.search import user_exists
from ocflib.account.search import user_is_group
from ocflib.account.search import user_is_sorried
from ocflib.account.search import users_by_callink_oid
from ocflib.account.search import users_by_calnet_uid
from ocflib.account.search import users_by_filter


class TestUsersByFilter:

    @pytest.mark.parametrize('filter_str,results', [
        ('(uid=ckuehl)', {'ckuehl'}),
        ('(uidNumber=28460)', {'ckuehl'}),
        ('(|(uidNumber=28460)(uid=daradib))', {'ckuehl', 'daradib'}),
        ('(uid=doesnotexist)', set()),
        ('(herp=derp)', set()),
        ('(!(uid=*))', set()),
    ])
    def test_users_by_filter(self, filter_str, results):
        assert set(users_by_filter(filter_str)) == results

    @pytest.mark.parametrize('filter_str', ['', 'uid=ckuehl', '42', 'asdf'])
    def test_invalid_filters(self, filter_str):
        with pytest.raises(Exception):
            users_by_filter(filter_str)


@pytest.mark.parametrize('uid,users', [
    (872544, ['daradib']),
    (666666, []),
])
def test_users_by_calnet_uid(uid, users):
    assert users_by_calnet_uid(uid) == users


@pytest.mark.parametrize('oid,users', [
    (46130, ['bpreview']),
    (666666, []),
])
def test_users_by_callink_oid(oid, users):
    assert users_by_callink_oid(oid) == users


class TestUserAttrs:

    def test_existing_user(self):
        user = user_attrs('ckuehl')
        assert user['uid'] == ['ckuehl']
        assert user['uidNumber'] == ['28460']

    def test_nonexistent_user(self):
        assert user_attrs('doesnotexist') is None


class TestUserAttrsUCB:

    def test_existing_user(self, test_uid=1034192):
        """These are a little flaky because alumni eventually get kicked out of
        the university's "People" OU. So you'll need to update these every few
        years."""
        user = user_attrs_ucb(test_uid)
        assert user['uid'] == [str(test_uid)]
        assert 'person' in user['objectClass']

    def test_nonexistent_user(self):
        assert user_attrs_ucb(9999999999) is None


@pytest.mark.parametrize('user,exists', [
    ('ckuehl', True),
    ('bpreview', True),
    ('doesnotexist', False),
])
def test_user_exists(user, exists):
    assert user_exists(user) == exists


@pytest.mark.parametrize('user,sorried', [
    ('ckuehl', False),
    ('sanjay', True),
])
def test_user_is_sorried(user, sorried):
    assert user_is_sorried(user) == sorried


class TestUserIsGroup:

    @pytest.mark.parametrize('user,exists', [
        ('ckuehl', False),
        ('bpreview', True),
    ])
    def test_user_is_group(self, user, exists):
        assert user_is_group(user) == exists

    def test_notexistent_user(self):
        with pytest.raises(Exception):
            user_is_group('doesnotexist')
